#!/usr/bin/env python3
"""
 Set odrive can parameters

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
from time import sleep

import odrive  # type: ignore
import odrive.enums as enums  # type: ignore
from utils import check_error


drv = odrive.find_any()
print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()
check_error(drv)

# disable watchdog
drv.axis0.config.enable_watchdog = False

ax = drv.axis0
ax.requested_state = enums.AxisState.FULL_CALIBRATION_SEQUENCE
sleep(1)
print("Calibrating...")
while drv.axis0.current_state != enums.AxisState.IDLE:
    sleep(1)
    print(".", end="", flush=True)

ax.motor.config.pre_calibrated = True
ax.encoder.config.pre_calibrated = True

try:
    check_error(drv)
    drv.save_configuration()
except Exception as e:
    print(e)

# drv.reboot()



#!/usr/bin/env python3
"""
 utils for odrive testing

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
from typing import List
import odrive.enums as enums


class OdriveError(Exception):
    """Custom exception for ODrive errors"""


def decode_error_flags(error_value: int, error_enum) -> List[str]:
    """
    Decode an error value into a list of human-readable error names
    by checking each bit against the enum values.
    """
    error_messages = []
    for error in error_enum:
        if error_value & error.value:
            error_messages.append(error.name)
    return error_messages


def check_error(drv) -> None:
    """
    Check if there are any errors and raise OdriveError with decoded error messages.
    Uses odrive.enums to decode error flags into human-readable messages.
    """
    ax = drv.axis0

    error_checks = [
        ("ODrive", ax.error, enums.ODriveError),
        ("Motor", ax.motor.error, enums.MotorError),
        ("Encoder", ax.encoder.error, enums.EncoderError),
        ("Controller", ax.controller.error, enums.ControllerError),
        (
            "Sensorless Estimator",
            ax.sensorless_estimator.error,
            enums.SensorlessEstimatorError,
        ),
    ]

    error_messages: List[str] = []

    for source, error_value, error_enum in error_checks:
        if error_value:
            decoded = decode_error_flags(error_value, error_enum)
            if decoded:
                error_messages.extend(f"{source}: {error}" for error in decoded)

    if error_messages:
        raise OdriveError("Multiple errors detected:\n" + "\n".join(error_messages))



#!/usr/bin/env python3
"""
 Demonstration of position control using CAN interface

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import time
import logging
from odrive_can.odrive import ODriveCAN, CanMsg
from odrive_can.tools import UDP_Client
from odrive_can.utils import run_main_async

SETTLE_TIME = 5.0  # settle time in [s]


log = logging.getLogger("pos_ctl")
udp = UDP_Client()


def feedback_callback(msg: CanMsg, caller: ODriveCAN):
    """position callback, send data to UDP client"""
    data = msg.data
    data["setpoint"] = caller.setpoint
    data["ts"] = time.time()
    udp.send({f"axis_{msg.axis_id}": data}, add_timestamp=False)


async def configure_controller(
    drv: ODriveCAN, input_mode: str = "POS_FILTER", accel: float = 120.0
):
    """setup control parameters"""
    log.info("Configuring controller")

    drv.clear_errors()

    # set parameters
    drv.set_pos_gain(5.0)

    drv.set_traj_vel_limit(40.0)
    drv.set_traj_accel_limits(accel, accel)

    # reset encoder
    drv.set_linear_count(0)

    drv.set_controller_mode("POSITION_CONTROL", input_mode)

    # set position control mode
    await drv.set_axis_state("CLOSED_LOOP_CONTROL")
    drv.check_errors()


async def main_loop(
    drv: ODriveCAN, input_mode: str = "POS_FILTER", amplitude: float = 40.0
):
    """position demo"""

    log.info("-----------Running position control-----------------")

    # start running

    drv.set_input_pos(amplitude)
    await asyncio.sleep(2)

    idx = 0
    try:
        while True:
            drv.check_errors()

            drv.set_input_pos(amplitude)
            idx += 1
            await asyncio.sleep(SETTLE_TIME)
            amplitude = -amplitude

    except KeyboardInterrupt:
        log.info("Stopping")
    finally:
        drv.stop()
        await asyncio.sleep(0.5)


async def main(
    axis_id: int,
    input_mode: str = "POS_FILTER",
    amplitude: float = 40.0,
):
    print("Starting position control demo, press CTRL+C to exit")
    log.info(f"{axis_id=} {input_mode=} {amplitude=}")
    drv = ODriveCAN(axis_id)
    drv.feedback_callback = feedback_callback

    await drv.start()
    await asyncio.sleep(0.5)
    # drv.reboot()
    # await asyncio.sleep(1.0)

    drv.check_alive()
    log.info("Clearing errors")
    drv.clear_errors()
    drv.check_errors()

    await configure_controller(drv, input_mode)
    await main_loop(drv, input_mode, amplitude)  # Added back the main loop


if __name__ == "__main__":
    run_main_async(main(1, "can0"))



#!/usr/bin/env python3
"""
 Set odrive can parameters

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import odrive  # type: ignore

drv = odrive.find_any()
print(f"Found ODrive {drv.serial_number}")


drv.axis0.config.can.node_id = 1
drv.axis0.config.can.encoder_rate_ms = 100
drv.axis0.config.can.heartbeat_rate_ms = 200


drv.axis1.config.can.node_id = 8
drv.axis1.config.can.encoder_rate_ms = 0
drv.axis1.config.can.heartbeat_rate_ms = 0

drv.can.config.baud_rate = 500000
drv.save_configuration()
drv.reboot()



#!/usr/bin/env python3
"""
 async can usage example

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import can

# pylint: disable=abstract-class-instantiated
BUS = can.interface.Bus(
    channel="vcan0", interface="socketcan", receive_own_messages=True
)


# Callback function to process received messages
def print_message(msg):
    print(f"callback: {msg}")


# Coroutine to send messages
async def send_messages():
    for i in range(10):
        message = can.Message(arbitration_id=i, data=[i] * 8)
        BUS.send(message)
        print(f"Sent: {message}")
        await asyncio.sleep(0.1)  # Delay to simulate processing

    raise StopIteration("done sending")


# Coroutine to receive messages
async def receive_messages(reader):
    while True:
        msg = await reader.get_message()
        print(f"awaited {msg}")


async def main():
    # Start the sending and receiving coroutines

    reader = can.AsyncBufferedReader()
    notifier = can.Notifier(BUS, [print_message, reader])

    try:
        await asyncio.gather(send_messages(), receive_messages(reader))
    except Exception as e:
        print(f"Exception: {e}")
    finally:
        print("Cleaning up")
        notifier.stop()  # Stop the notifier when done
        BUS.shutdown()


# Run the main coroutine
if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python3
"""
 support functions

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import asyncio
import logging
import os
from functools import wraps
from pathlib import Path
from typing import Any, Coroutine

import can
import cantools
import coloredlogs  # type: ignore

LOG_FORMAT = "%(asctime)s [%(name)s] %(filename)s:%(lineno)d - %(message)s"
TIME_FORMAT = "%H:%M:%S.%f"


def get_axis_id(msg: can.Message) -> int:
    """get axis id from message"""
    return msg.arbitration_id >> 5


def extract_ids(can_id: int) -> tuple[int, int]:
    """get axis_id and cmd_id from can_id"""
    cmd_id = can_id & 0x1F  # Extract lower 5 bits for cmd_id
    axis_id = can_id >> 5  # Shift right by 5 bits to get axis_id
    return axis_id, cmd_id


# pylint: disable=import-outside-toplevel
def get_dbc(name: str = "odrive-cansimple-0.5.6"):
    """get the cantools database"""

    # get relative path to db file
    dbc_path = Path(__file__).parent / f"dbc/{name}.dbc"

    return cantools.database.load_file(dbc_path.as_posix())


def get_dbc_path(name: str = "odrive-cansimple-0.5.6") -> Path:
    """get the path to the dbc file"""
    return Path(__file__).parent / f"dbc/{name}.dbc"


def run_main_async(
    coro: Coroutine[Any, Any, None], silence_loggers: list[str] | None = None
) -> None:
    """convenience function to avoid code duplication"""
    loglevel = os.environ.get("LOGLEVEL", "INFO").upper()
    coloredlogs.install(level=loglevel, fmt=LOG_FORMAT)
    logging.info(f"Log level set to {loglevel}")

    if silence_loggers:
        for logger in silence_loggers:
            logging.info(f"Silencing logger: {logger}")
            logging.getLogger(logger).setLevel(logging.CRITICAL)

    try:
        asyncio.run(coro)
    except KeyboardInterrupt:
        pass
    except ExceptionGroup as group:
        logging.error("ExceptionGroup caught")
        for e in group.exceptions:  # pylint: disable=not-an-iterable
            logging.exception(f"Caught exception: {e}", exc_info=e)
    except asyncio.CancelledError:
        logging.error("Cancelled")

    except Exception as e:
        logging.error(f"Crashed with {e},  type: {type(e)}", exc_info=e)


def async_timeout(timeout: float = 1.0):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await asyncio.wait_for(func(*args, **kwargs), timeout)
            except asyncio.TimeoutError:
                func_name = func.__name__
                # Include the function name in the error message
                raise TimeoutError(f"{func_name} timed out after {timeout} seconds")

        return wrapper

    return decorator



#!/usr/bin/env python3
"""
 Watchdog demo.

 1. Set costant velocity and continue sending setpoints
 2. pause sending for 3 seconds
 3. check for watchdog, reset if needed and resume sending setpoints

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

from odrive_can.odrive import ODriveCAN

from odrive_can.tools import UDP_Client
from odrive_can.utils import run_main_async

from odrive_can.examples.position_control import feedback_callback

VERSION = "2024.02.17"

SETPOINT_DELAY = 0.1


log = logging.getLogger("watchdog")
udp = UDP_Client()

setpoint_enabled = True


async def configure_controller(drv: ODriveCAN):
    """setup control parameters"""

    log.info("Configuring controller")

    drv.clear_errors()

    # reset encoder
    drv.set_linear_count(0)

    drv.set_controller_mode("VELOCITY_CONTROL", "VEL_RAMP")

    # set position control mode
    await drv.wait_for_heartbeat()
    drv.check_errors()
    await drv.set_axis_state("CLOSED_LOOP_CONTROL")
    # await drv.wait_for_heartbeat()


async def setpoint_loop(drv: ODriveCAN, setpoint: float):
    """send setpoints if enabled"""

    log.info("Sending setpoints")
    while True:
        drv.check_errors()

        if setpoint_enabled:
            drv.set_input_vel(setpoint)

        await asyncio.sleep(SETPOINT_DELAY)


async def setpoint_enabler():
    """enable setpoints after 3 seconds"""

    global setpoint_enabled

    while True:
        log.info("Enabling setpoints")
        setpoint_enabled = True
        await asyncio.sleep(3)
        log.info("Disabling setpoints")
        setpoint_enabled = False
        await asyncio.sleep(3)


async def main(axis_id: int, amplitude: float = 40.0):
    """velocity control demo"""
    global setpoint_enabled

    log.info(f"Running watchdog demo on axis {axis_id} ")

    drv = ODriveCAN(axis_id)

    drv.feedback_callback = feedback_callback
    await drv.start()

    await configure_controller(drv)

    try:
        async with asyncio.TaskGroup() as tg:
            tg.create_task(setpoint_loop(drv, amplitude))
            tg.create_task(setpoint_enabler())
    except ExceptionGroup as e:  # Catch the ExceptionGroup
        for exception in e.exceptions:  #  pylint: disable=not-an-iterable
            log.error(exception)
    finally:
        drv.stop()
        await asyncio.sleep(0.5)


if __name__ == "__main__":
    run_main_async(main(1))



#!/usr/bin/env python3
"""
 inspect and decode CAN messages for ODrive

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import can
import odrive_can


TIMEOUT = 1.0


def receive_and_decode(bus: can.Bus, dbc):
    while True:
        message = bus.recv(TIMEOUT)
        if message is None:
            raise TimeoutError("Timeout occurred, no message.")
        try:
            # Attempt to decode the message using the DBC file
            decoded_message = dbc.decode_message(message.arbitration_id, message.data)
            print(f"Received Message: {decoded_message}")
        except KeyError:
            # If the message ID is not in the DBC file, print the raw message
            print(f"Received Raw Message: {message}")


def main():
    # Load the DBC file
    dbc = odrive_can.get_dbc()

    channel = "vcan0"  # Change to your CAN interface
    bus = can.Bus(channel=channel, interface="socketcan", receive_own_messages=True)

    try:
        receive_and_decode(bus, dbc)
    except KeyboardInterrupt:
        print("Stopped")
    except TimeoutError as error:
        print(error)
    finally:
        bus.shutdown()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=broad-except, redefined-outer-name
import asyncio
import logging
import argparse
import coloredlogs  # type: ignore

from odrive_can import LOG_FORMAT, TIME_FORMAT
from odrive_can.odrive import ODriveCAN
from odrive_can.timer import timeit

log = logging.getLogger()
coloredlogs.install(level="INFO", fmt=LOG_FORMAT, datefmt=TIME_FORMAT)

# set can logger to INFO
logger_can = logging.getLogger("can").setLevel(logging.INFO)  # type: ignore


def parse_args():
    parser = argparse.ArgumentParser(description="ODrive CAN test script")
    parser.add_argument("--axis-id", type=int, default=0, help="ODrive axis ID")
    parser.add_argument("--interface", type=str, default="vcan0", help="CAN interface")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    return parser.parse_args()


@timeit
async def request(drv: ODriveCAN, method: str):
    """request data from ODrive"""
    log.info(f"Requesting {method}")
    fcn = getattr(drv, method)
    data = await fcn()
    log.info(f"{data=}")


async def main(args):
    if args.debug:
        coloredlogs.set_level("DEBUG")

    try:
        drv = ODriveCAN(axis_id=args.axis_id)
        await drv.start()

        drv.check_alive()
        drv.check_errors()

        for param in [
            "get_bus_voltage_current",
            "get_motor_error",
            "get_encoder_error",
            "get_sensorless_error",
            "get_encoder_estimates",
            "get_encoder_count",
            "get_iq",
            "get_sensorless_estimates",
            "get_adc_voltage",
            "get_controller_error",
        ]:
            await request(drv, param)

        # set velocity control mode
        log.info("Setting velocity control mode")
        await drv.set_axis_state("CLOSED_LOOP_CONTROL")
        log.info(f"Currrent axis state: {drv.axis_state}")
        await asyncio.sleep(1)

    except Exception as e:  # pylint: disable=broad-except
        log.error(e)
    finally:
        drv.stop()
        await asyncio.sleep(0.5)


if __name__ == "__main__":
    args = parse_args()
    try:
        asyncio.run(main(args))
    except KeyboardInterrupt:
        log.info("KeyboardInterrupt")
    except Exception as e:
        log.error(e)



#!/usr/bin/env python3
"""
 Send some test messages to the drive

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
from typing import Optional
from time import sleep
import can
from odrive_can import get_dbc, get_axis_id

# pylint: disable=abstract-class-instantiated
BUS = can.interface.Bus(
    channel="vcan0", interface="socketcan", receive_own_messages=False
)
DB = get_dbc()  # Load the DBC file


# Callback function to process received messages
def print_message(msg):
    """print message, if not ignored"""

    axis_id = get_axis_id(msg)

    try:
        # Attempt to decode the message using the DBC file
        decoded_message = DB.decode_message(msg.arbitration_id, msg.data)
        print(f"Decoded {axis_id=}: {decoded_message}")
    except KeyError:
        # If the message ID is not in the DBC file, print the raw message
        print(f"Raw {axis_id=}: {msg}")


def send_message(msg_name: str, msg_dict: Optional[dict] = None, rtr: bool = False):
    """send message by name. If no msg_dict is provided, use zeros"""
    msg = DB.get_message_by_name(msg_name)
    if rtr:
        # For RTR messages, don't specify the data field
        msg = can.Message(
            arbitration_id=msg.frame_id,
            is_extended_id=False,
            is_remote_frame=True,
        )
    else:
        if msg_dict is None:
            msg_dict = {signal.name: 0 for signal in msg.signals}
        data = msg.encode(msg_dict)
        msg = can.Message(
            arbitration_id=msg.frame_id,
            data=data,
            is_extended_id=False,
        )

    BUS.send(msg)


def send_messages():
    for axis_id in [0, 1]:
        for cmd_name in [
            "Get_Bus_Voltage_Current",
            "Get_Motor_Error",
            "Get_Encoder_Error",
            "Get_Encoder_Estimates",
            "Get_Iq",
            "Get_Sensorless_Estimates",
            "Clear_Errors",
        ]:
            msg_name = f"Axis{axis_id}_{cmd_name}"
            print(f"Sending {msg_name}")
            send_message(msg_name, rtr=True)
            sleep(0.1)


def main():
    # Start the sending and receiving coroutines

    notifier = can.Notifier(BUS, [print_message])

    # Send some messages
    send_messages()

    print("Cleaning up")
    notifier.stop()  # Stop the notifier when done
    BUS.shutdown()


# Run the main coroutine
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python3
"""
 Generate interface from dbc file

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

from pathlib import Path
import cantools
from odrive_can import get_dbc

script_name = Path(__file__).name

HEADER = """#!/usr/bin/env python3
## interface generated by {script_name}

from abc import ABC, abstractmethod

class DbcInterface(ABC):
    @abstractmethod
    async def request(self, msg_name: str, timeout: float = 0.5) -> dict:
        pass
"""


GET_TEMPLATE = """
    async def {method_name}(self) -> dict:
        return await self.request("{param}")\n"""


def split_message_name(message_name: str) -> tuple[str, str, str | None]:
    """split to axis, method_name, prefix (get/set)"""
    axis, method_name = message_name.split("_", 1)

    if method_name.lower().startswith("get"):
        prefix = "get"

    elif method_name.lower().startswith("set"):
        prefix = "set"

    else:
        prefix = None

    return axis, method_name, prefix


def get_method(message: cantools.db.Message) -> str:
    """get method string"""
    _, method_name, _ = split_message_name(message.name)

    code = GET_TEMPLATE.format(method_name=method_name.lower(), param=method_name)

    return code


def set_method(message: cantools.db.Message) -> str:
    """Generate a set method string from a CAN message."""
    _, method_name, _ = split_message_name(message.name)

    # Extracting signal names and types
    params = []
    param_dict = {}
    for signal in message.signals:
        signal_name = signal.name
        param_type = "int"
        if signal.is_float:
            param_type = "float"
        elif isinstance(signal.choices, dict):
            param_type = "str"  # Enumerated signal

        params.append(f"{signal_name.lower()}: {param_type}")
        param_dict[signal_name] = signal_name.lower()

    params_str = ", ".join(params)
    params_dict_str = ", ".join([f'"{k}": {v}' for k, v in param_dict.items()])

    # Template for the set method
    set_template = f"""
    def {method_name.lower()}(self, {params_str}):
        self._send_message("{method_name}", {{{{{params_dict_str}}}}})\n"""

    return set_template.format(
        method_name=method_name,
        message_name=message.name,
    )


def dbc_to_code(db, dest):
    """create interface from dbc database"""

    get_methods = []
    set_methods = []

    for message in db.messages[:28]:
        # split message name by first underscore

        print(f"Processing {message.name}")
        _, _, prefix = split_message_name(message.name)

        if prefix == "get":
            code = get_method(message)
            get_methods.append(code)

        elif prefix == "set":
            code = set_method(message)
            set_methods.append(code)

    print(f"Generating {dest}")
    # Open the Markdown file for writing
    with open(dest, "w", encoding="utf8") as output_file:
        # Write a header
        code = HEADER.format(script_name=script_name)
        output_file.write(code)

        for method in get_methods:
            output_file.write(method)

        for method in set_methods:
            output_file.write(method)


db = get_dbc("odrive-cansimple-0.5.6")
dbc_to_code(db, "interface.py")



import asyncio
from contextlib import asynccontextmanager
import os
import pytest
from odrive_can.mock import ODriveCANMock

from odrive_can.odrive import ODriveCAN, CanMsg


AXIS_ID = 1

# enable tests if vcan0 is available
if not os.path.exists("/sys/class/net/vcan0"):
    pytestmark = pytest.mark.skip(reason="No vcan0 present.")


class Feedback:
    """feedback collection class"""

    def __init__(self):
        self.pos = 0
        self.vel = 0

    def feedback_callback(self, msg: CanMsg, caller: ODriveCAN):
        data = msg.data
        self.pos = data["Pos_Estimate"]
        self.vel = data["Vel_Estimate"]


# tried to use a fixture, but it did not work. so using a context manager
@asynccontextmanager
async def odrive_can_context(axis_id=AXIS_ID):
    odrv = ODriveCANMock(axis_id=axis_id)
    odrv_task = asyncio.create_task(odrv.main())
    await asyncio.sleep(1)  # Allow some time for the mock to initialize
    try:
        yield odrv
    finally:
        odrv_task.cancel()
        try:
            await odrv_task
        except asyncio.CancelledError:
            pass


@pytest.mark.asyncio
async def test_velocity_control():
    async with odrive_can_context():
        fbk = Feedback()

        drv = ODriveCAN(axis_id=AXIS_ID)
        drv.feedback_callback = fbk.feedback_callback
        await drv.start()

        drv.check_alive()
        # reset encoder
        drv.set_linear_count(0)

        drv.set_controller_mode("VELOCITY_CONTROL", "VEL_RAMP")

        # set position control mode
        await drv.set_axis_state("CLOSED_LOOP_CONTROL")
        drv.check_errors()

        setpoint = 5
        drv.set_input_vel(setpoint)
        await asyncio.sleep(1)
        assert fbk.vel == setpoint



#!/usr/bin/env python3
"""
 simple test of velocity control

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import asyncio
import os
import odrive  # type: ignore
import odrive.enums as enums  # type: ignore
from utils import check_error

from odrive_can.tools import UDP_Client

SETPOINT = 50
DURATION = 20

drv = odrive.find_any()

print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()
check_error(drv)

ax = drv.axis0


ax.controller.config.control_mode = enums.ControlMode.VELOCITY_CONTROL
ax.controller.config.input_mode = enums.InputMode.VEL_RAMP
ax.requested_state = enums.AxisState.CLOSED_LOOP_CONTROL

ax.config.enable_watchdog = False

# limit acceleration
ax.controller.config.vel_ramp_rate = 60
ax.motor.config.current_lim = 20

# set positiion to zero
ax.encoder.set_linear_count(0)

# disable limits
ax.controller.config.enable_vel_limit = True
ax.controller.config.vel_limit = 100


# ignore hall errors
ax.encoder.config.ignore_illegal_hall_state = False


async def feedback_loop() -> None:
    udp_dest = os.getenv("UDP_DATA_DEST", "localhost")

    print(f"sending data to {udp_dest}")
    udp = UDP_Client(host=udp_dest)

    counter = 0

    while True:
        check_error(drv)

        sp_rpm = int(ax.controller.input_vel * 60)
        vel_rpm = int(ax.encoder.vel_estimate * 60)

        data = {
            "sp": sp_rpm,
            "vel": vel_rpm,
            "pos": ax.encoder.pos_estimate,
        }

        udp.send(data)
        await asyncio.sleep(0.05)

        if counter % 100 == 0:
            print(f"sp: {data['sp']}, vel: {data['vel']}")

        counter += 1


async def setpoint_loop() -> None:
    print("Starting setpoint loop")
    ax.controller.input_vel = 0
    await asyncio.sleep(1)
    while True:
        ax.controller.input_vel = SETPOINT
        print(f"Setpoint set to {ax.controller.input_vel} ({SETPOINT*60} RPM)")
        await asyncio.sleep(DURATION)

        ax.controller.input_vel = 0
        print(f"Setpoint set to {ax.controller.input_vel}")
        await asyncio.sleep(5)


async def main() -> None:
    await asyncio.gather(feedback_loop(), setpoint_loop())


try:
    asyncio.run(main())
except KeyboardInterrupt:
    pass
finally:
    ax.controller.input_vel = 0
    ax.requested_state = enums.AxisState.IDLE



#!/usr/bin/env python3
"""
 demonstrate usage of OdriveCAN class

 set env variables first:
   "CAN_CHANNEL": "can0",
   "CAN_INTERFACE": "socketcan",

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

import coloredlogs  # type: ignore

from odrive_can import LOG_FORMAT, TIME_FORMAT
from odrive_can.odrive import CommandId, ODriveCAN
from odrive_can.tools import UDP_Client
from odrive_can.timer import timeit

log = logging.getLogger()
coloredlogs.install(level="DEBUG", fmt=LOG_FORMAT, datefmt=TIME_FORMAT)


udp = UDP_Client()

AXIS_ID = 1


def position_callback(data):
    """position callback, send data to UDP client"""
    udp.send(data)


async def get_bus_voltage_current(drv: ODriveCAN):
    """request bus voltage and current"""

    @timeit
    async def request(drv: ODriveCAN):
        data = await drv.get_bus_voltage_current()
        log.info(f"{data=}")

    log.info("Requesting bus voltage and current")
    # request bus voltage and current
    for _ in range(4):
        log.info("------------------------------")
        try:
            await request(drv)

        except Exception as e:  # pylint: disable=broad-except
            log.warning(e)
        await asyncio.sleep(0.1)


async def position_control(drv: ODriveCAN):
    """simple position control loop"""

    log.info("-----------Running position control-----------------")

    # set positiion to zero
    drv.set_linear_count(0)

    drv.set_controller_mode("POSITION_CONTROL", "POS_FILTER")
    drv.set_pos_gain(3.0)

    # setpoint
    setpoint = 20
    duration = 5

    for _ in range(4):
        try:
            drv.check_errors()
            log.info(f"Setting position setpoint to {setpoint}")
            drv.set_input_pos(setpoint)
            await asyncio.sleep(duration)
            log.info(f"Setting position setpoint to {-setpoint}")
            drv.set_input_pos(-setpoint)
            await asyncio.sleep(duration)

        except Exception as e:  # pylint: disable=broad-except
            log.warning(e)


async def main():
    drv = ODriveCAN(axis_id=AXIS_ID)
    drv.feedback_callback = position_callback
    await drv.start()

    # log some messages
    await asyncio.sleep(1.0)

    await drv.set_axis_state("CLOSED_LOOP_CONTROL")

    # set log level to INFO
    coloredlogs.set_level(logging.INFO)

    # ignore encoder estimate messages
    drv.ignore_message(CommandId.ENCODER_ESTIMATE)

    # request bus voltage and current
    await get_bus_voltage_current(drv)

    # enable encoder feedback
    drv.allow_message(CommandId.ENCODER_ESTIMATE)

    # reset encoder
    drv.set_linear_count(0)
    await asyncio.sleep(1.0)

    # run position control
    await position_control(drv)

    # shutdown
    drv.stop()
    await asyncio.sleep(0.5)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3
"""
 Demonstration of velocity control using CAN interface

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

from odrive_can.odrive import ODriveCAN
from odrive_can.setpoints import sawtooth_generator
from odrive_can.tools import UDP_Client
from odrive_can.utils import run_main_async

from odrive_can.examples.position_control import feedback_callback

VERSION = "2024.02.17"

SETPOINT_DELAY = 0.1


log = logging.getLogger("pos_ctl")
udp = UDP_Client()


async def configure_controller(drv: ODriveCAN):
    """setup control parameters"""

    # reset encoder
    drv.set_linear_count(0)

    drv.set_controller_mode("VELOCITY_CONTROL", "VEL_RAMP")

    # set position control mode
    await drv.set_axis_state("CLOSED_LOOP_CONTROL")
    await drv.wait_for_heartbeat()
    drv.check_errors()


async def main(axis_id: int, amplitude: float = 40.0):
    """velocity control demo"""

    log.info(f"Running velocity control demo on axis {axis_id} ")

    drv = ODriveCAN(axis_id)

    drv.feedback_callback = feedback_callback
    await drv.start()

    await asyncio.sleep(0.5)
    drv.check_alive()
    log.info("Clearing errors")
    drv.clear_errors()
    await drv.wait_for_heartbeat()
    drv.check_errors()

    await configure_controller(drv)

    # make setpoint generator
    setpoint_gen = sawtooth_generator(roc=10.0, max_val=40.0)

    log.info("Running velocity control")
    try:
        while True:
            drv.check_errors()
            setpoint = next(setpoint_gen)

            drv.set_input_vel(setpoint)
            await asyncio.sleep(SETPOINT_DELAY)

    finally:
        drv.stop()
        await asyncio.sleep(0.5)


if __name__ == "__main__":
    run_main_async(main(1))



#!/usr/bin/env python3
"""
 simple test of velocity control

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import time

import odrive  # type: ignore
import odrive.enums as enums  # type: ignore
from utils import check_error

from odrive_can.tools import UDP_Client

# send data to plotjuggler
udp = UDP_Client(host="localhost")

drv = odrive.find_any()


print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()
drv.axis0.controller.config.control_mode = enums.ControlMode.VELOCITY_CONTROL
drv.axis0.requested_state = enums.AxisState.CLOSED_LOOP_CONTROL
drv.axis0.config.enable_watchdog = False

# limit acceleration
drv.axis0.controller.config.vel_ramp_rate = 10


# set positiion to zero
drv.axis0.encoder.set_linear_count(0)


SMOOTH_PROFILE = (
    list(range(0, 50))
    + 20 * [50]
    + list(range(50, -50, -1))
    + 20 * [-50]
    + list(range(-50, 0))
)


def follow_curve(setpoints, delay=0.02):
    """follow curve consisting of velocity setpoints"""

    for setpoint in setpoints:
        check_error(drv)

        drv.axis0.controller.input_vel = setpoint

        data = {
            "ts": time.time(),
            "sp": setpoint,
            "vel": drv.axis0.encoder.vel_estimate,
            "pos": drv.axis0.encoder.pos_estimate,
        }

        udp.send(data)
        time.sleep(delay)


try:
    print("---------------------Smooth profile---------------------")
    i = 0
    while True:
        print(f"run {i}")
        follow_curve(SMOOTH_PROFILE)
        i += 1


except KeyboardInterrupt:
    print("interrupted")
finally:
    drv.axis0.controller.input_vel = 0
    drv.axis0.requested_state = enums.AxisState.IDLE



# type: ignore
import os
import time
import re

from click import prompt
from invoke import task


def get_version_from_dist():
    dist_files = os.listdir("dist")
    for filename in dist_files:
        match = re.match(r"beta_main-(\d+\.\d+\.\d+(\.dev\d+)?)", filename)
        if match:
            return match.group(1)
    raise ValueError("Version not found in dist folder")


@task
def clean(ctx):
    """
    Remove all files and directories that are not under version control to ensure a pristine working environment.
    Use caution as this operation cannot be undone and might remove untracked files.

    """

    ctx.run("git clean -nfdx")

    if (
        prompt(
            "Are you sure you want to remove all untracked files? (y/n)", default="n"
        )
        == "y"
    ):
        ctx.run("git clean -fdx")


@task
def lint(ctx):
    """
    Perform static analysis on the source code to check for syntax errors and enforce style consistency.
    """
    ctx.run("pylint -E src")
    ctx.run("mypy src")


@task
def test(ctx):
    """
    Run tests with coverage reporting to ensure code functionality and quality.
    """
    ctx.run("pytest --cov=src --cov-report term-missing tests", pty=True)


@task
def ci(ctx):
    """
    run ci locally in a fresh container

    """
    t_start = time.time()
    # get script directory
    script_dir = os.path.dirname(os.path.realpath(__file__))
    try:
        ctx.run(f"docker run --rm -v {script_dir}:/workspace roxauto/python-ci")
    finally:
        t_end = time.time()
        print(f"CI run took {t_end - t_start:.1f} seconds")


@task
def build_package(ctx):
    """
    Build package in docker container.
    """

    ctx.run("rm -rf dist")
    t_start = time.time()
    # get script directory
    script_dir = os.path.dirname(os.path.realpath(__file__))
    try:
        ctx.run(
            f"docker run --rm -v {script_dir}:/workspace roxauto/python-ci /scripts/build.sh"
        )
    finally:
        t_end = time.time()
        print(f"CI run took {t_end - t_start:.1f} seconds")


@task
def release(ctx):
    """publish package to pypi"""
    script_dir = os.path.dirname(os.path.realpath(__file__))

    token = os.getenv("PYPI_TOKEN")
    if not token:
        raise ValueError("PYPI_TOKEN environment variable is not set")

    ctx.run(
        f"docker run --rm -e PYPI_TOKEN={token} -v {script_dir}:/workspace roxauto/python-ci /scripts/publish.sh"
    )



from importlib.metadata import version, PackageNotFoundError
from .utils import get_axis_id, get_dbc, extract_ids, LOG_FORMAT, TIME_FORMAT
from .odrive import ODriveCAN, CanMsg

# Specify the public API of the module
__all__ = [
    "get_axis_id",
    "get_dbc",
    "extract_ids",
    "LOG_FORMAT",
    "TIME_FORMAT",
    "ODriveCAN",
    "CanMsg",
]

try:
    __version__ = version("odrive_can")
except PackageNotFoundError:
    # Package is not installed, and therefore, version is unknown.
    __version__ = "0.0.0+unknown"



import can
import pytest

from odrive_can.odrive import CanMsg, DriveError, HeartbeatError, ODriveCAN, dbc

# pylint: disable=redefined-outer-name, protected-access


# Mocking the dbc.get_message_by_frame_id and dbc.get_message_by_name methods
@pytest.fixture(autouse=True)
def mock_dbc_methods(mocker):
    mocker.patch.object(dbc, "get_message_by_frame_id", autospec=True)
    mocker.patch.object(dbc, "get_message_by_name", autospec=True)
    # Add any necessary logic for the mocked methods if needed


@pytest.fixture
def mock_can_message():
    return can.Message(arbitration_id=123, data=[0, 1, 2, 3, 4, 5, 6, 7])


@pytest.fixture
def odrive_can(mocker):
    mocker.patch("odrive_can.odrive.can.interface.Bus")
    return ODriveCAN(axis_id=0)


def test_can_msg_init(mock_can_message):
    can_msg = CanMsg(mock_can_message)
    assert can_msg.name  # Add more specific assertions based on your dbc parsing logic
    _ = str(can_msg)  # Test __str__ method


def test_check_alive_no_heartbeat(odrive_can):
    with pytest.raises(HeartbeatError):
        odrive_can.check_alive()


def test_check_alive_with_heartbeat(odrive_can, mock_can_message):
    # Mocking a heartbeat message
    mock_can_message.arbitration_id = 1  # Arbitration ID for heartbeat
    can_msg = CanMsg(mock_can_message)
    odrive_can._last_heartbeat = can_msg

    try:
        odrive_can.check_alive()
    except HeartbeatError:
        pytest.fail("HeartbeatError raised unexpectedly")


# test check errors
def test_check_errors_no_errors(odrive_can, mock_can_message):
    mock_can_message.arbitration_id = 1  # Arbitration ID for heartbeat
    heartbeat_msg = CanMsg(mock_can_message)
    heartbeat_msg.data = {
        "Axis_Error": "NONE",
        "Motor_Error_Flag": 0,
        "Encoder_Error_Flag": 0,
        "Controller_Error_Flag": 0,
    }
    odrive_can._last_heartbeat = heartbeat_msg

    try:
        odrive_can.check_errors()
    except DriveError:
        pytest.fail("DriveError raised unexpectedly")


def test_check_errors_with_errors(odrive_can, mock_can_message):
    mock_can_message.arbitration_id = 1  # Arbitration ID for heartbeat
    heartbeat_msg = CanMsg(mock_can_message)
    heartbeat_msg.data = {
        "Axis_Error": "ERROR",
        "Motor_Error_Flag": 1,
        "Encoder_Error_Flag": 0,
        "Controller_Error_Flag": 0,
    }
    odrive_can._last_heartbeat = heartbeat_msg

    with pytest.raises(DriveError):
        odrive_can.check_errors()



from click.testing import CliRunner
from odrive_can.cli import cli


def test_cli():
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Usage: cli [OPTIONS]" in result.output
    assert "Options:" in result.output
    assert "--help  Show this message and exit." in result.output


def test_cli_info():
    runner = CliRunner()
    result = runner.invoke(cli, ["info"])
    assert result.exit_code == 0



#!/usr/bin/env python3
"""
 Home odrive

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
from time import sleep

import odrive  # type: ignore
import odrive.enums as enums  # type: ignore

from utils import check_error


drv = odrive.find_any()
print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()
check_error(drv)

ax = drv.axis0

# enable endstop
ax.min_endstop.config.enabled = True
ax.min_endstop.config.is_active_high = True
ax.min_endstop.config.offset = 25.0

ax.requested_state = enums.AxisState.HOMING
sleep(1)
print("Homing...")
while ax.current_state != enums.AxisState.IDLE:
    sleep(1)
    print(".", end="", flush=True)

try:
    check_error(drv)
except Exception as e:
    print(e)
    print("Homing failed")
    exit(1)

print("Homing OK")

# get current position
pos = ax.encoder.pos_estimate
print(f"Current position: {pos}")



#!/usr/bin/env python3
"""
 Example of setpoint generation using linear model

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
from odrive_can.tools import UDP_Client
from odrive_can.setpoints import sawtooth_generator


DELAY = 0.1


async def main():
    """main"""

    udp = UDP_Client()

    generator = sawtooth_generator(roc=5.0)

    while True:
        setpoint = next(generator)
        udp.send({"setpoint": setpoint})
        await asyncio.sleep(DELAY)
        # print(".", end="", flush=True)


try:
    asyncio.run(main())
except KeyboardInterrupt:
    print("\nexiting")



#!/usr/bin/env python3
"""
 simulate CAN messages for ODrive

 1. setup virtual CAN interface on host. Use `install/setup_vcan_service.sh` script for that.
 2. terminal1: `candump vcan0`
 3. terminal2: `python3 sim_can.py`

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""


import asyncio
import can

import odrive_can

# pylint: disable=abstract-class-instantiated
BUS = can.interface.Bus(channel="vcan0", interface="socketcan")


async def send_heartbeat(dbc, axis_id):
    print("Starting heartbeat loop")

    # Fetch the "Axis0_Heartbeat" message from the DBC database
    heartbeat_msg = dbc.get_message_by_name(f"Axis{axis_id}_Heartbeat")

    while True:
        # Construct the data payload using the DBC message definition
        data = heartbeat_msg.encode(
            {
                "Axis_Error": 0,
                "Axis_State": 0,
                "Motor_Error_Flag": 0,
                "Encoder_Error_Flag": 0,
                "Controller_Error_Flag": 0,
                "Trajectory_Done_Flag": 0,
            }
        )

        # Send the message
        message = can.Message(
            arbitration_id=heartbeat_msg.frame_id, data=data, is_extended_id=False
        )
        BUS.send(message)

        await asyncio.sleep(1)


async def encoder_loop(dbc, axis_id, delay=0.1):
    print("Starting encoder loop")
    position = 0.0
    while True:
        encoder_msg = dbc.get_message_by_name(f"Axis{axis_id}_Get_Encoder_Estimates")
        data = encoder_msg.encode({"Pos_Estimate": position, "Vel_Estimate": 0.0})
        message = can.Message(
            arbitration_id=encoder_msg.frame_id, data=data, is_extended_id=False
        )
        BUS.send(message)
        position += 0.01  # Increment position to simulate movement
        await asyncio.sleep(delay)


async def main():
    # Load the DBC file
    dbc = odrive_can.get_dbc()

    axis_id = 0  # Change to your desired axis ID

    async with asyncio.TaskGroup() as tg:
        tg.create_task(send_heartbeat(dbc, axis_id))
        tg.create_task(encoder_loop(dbc, axis_id))


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("Stopped")
    finally:
        BUS.shutdown()



#!/usr/bin/env python3
"""
 mock ODrive CAN interface

 This is just a rough mock, not a complete simulation. This model uses
 infinite acceleration for example

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging
from typing import Optional

import can
from can.interfaces.socketcan import SocketcanBus
from can.interfaces.udp_multicast import UdpMulticastBus
import coloredlogs  # type: ignore

from odrive_can import LOG_FORMAT, TIME_FORMAT, get_axis_id, get_dbc
from odrive_can.can_utils import get_can_bus
from odrive_can.linear_model import LinearModel

# pylint: disable=abstract-class-instantiated, unnecessary-lambda, broad-except

# set can logger to INFO
logger_can = logging.getLogger("can").setLevel(logging.INFO)  # type: ignore


class OdriveMock:
    """mock physical ODrive device, excluding CAN interface"""

    def __init__(self, roc: float = 10.0, logger: Optional[logging.Logger] = None):
        """mock ODrive device

        Args:
            roc (float, optional): rate of change (velocity of poisition). Defaults to 10.0.
            logger (Optional[logging.Logger], optional): logger object. Defaults to None.
        """
        self.log = logger or logging.getLogger("odrive_mock")
        self.model = LinearModel(roc=roc)

        self.axis_state = "IDLE"
        self.input_mode = "INACTIVE"
        self.control_mode = "VELOCITY_CONTROL"

        self._accum_pos = 0.0  # accumulated position, when in velocity mode
        self._accum_pos_prev = 0.0  # previous accumulated position

        # velocity estimate when in position mode
        self._velocity_estimate = 0.0

    async def set_axis_state(self, data: dict):
        """set axis state"""
        state = data["Axis_Requested_State"]
        self.log.info(f"Setting axis state to {state}")
        await asyncio.sleep(0.5)
        self.axis_state = state
        self.log.info(f"Axis state is now {self.axis_state}")

    def set_controller_mode(self, data: dict):
        """set controller mode"""
        self.log.info(f"Setting controller mode to {data}")
        self.control_mode = data["Control_Mode"]
        self.input_mode = data["Input_Mode"]

    def set_input_pos(self, pos: float):
        """position setpoint"""
        if self.control_mode != "POSITION_CONTROL":
            self.log.warning("Ignoring setpoint. Not in position control mode")
            return
        self.log.debug(f"Setting input pos to {pos}")
        self.model.setpoint = pos

    def set_input_vel(self, vel: float):
        """velocity setpoint"""
        if self.control_mode != "VELOCITY_CONTROL":
            self.log.warning("Ignoring setpoint. Not in velocity control mode")
            return
        self.log.debug(f"Setting input vel to {vel}")
        self.model.setpoint = vel

    @property
    def position(self) -> float:
        """position"""
        if self.control_mode == "POSITION_CONTROL":
            return self.model.val
        elif self.control_mode == "VELOCITY_CONTROL":
            return self._accum_pos
        else:
            return 0.0

    @property
    def velocity(self) -> float:
        """velocity"""
        if self.control_mode == "POSITION_CONTROL":
            return self._velocity_estimate
        elif self.control_mode == "VELOCITY_CONTROL":
            return self.model.val
        else:
            return 0.0

    def update(self):
        """update model"""
        dt = self.model.step()
        if self.control_mode == "VELOCITY_CONTROL":
            self._accum_pos += self.model.val * dt
        elif self.control_mode == "POSITION_CONTROL":
            self._accum_pos = self.model.val

        self._velocity_estimate = (self._accum_pos - self._accum_pos_prev) / dt
        self._accum_pos_prev = self._accum_pos


class ODriveCANMock:
    """class to mock ODrive CAN interface"""

    def __init__(
        self,
        axis_id: int = 0,
        can_bus: UdpMulticastBus | SocketcanBus | None = None,
    ):
        self.log = logging.getLogger(f"odrive.mock.{axis_id}")

        self.log.info(f"Starting mock {axis_id=}")
        self.dbc = get_dbc()
        self.axis_id = axis_id

        self.bus = can_bus or get_can_bus()
        self.can_reader = can.AsyncBufferedReader()
        self.notifier = can.Notifier(self.bus, [self.can_reader])

        self.odrive = OdriveMock(logger=self.log)


    def set_rate_of_change(self, roc: float) -> None:
        """set rate of change"""
        self.odrive.model.roc = roc

    async def message_handler(self):
        """handle received message"""

        self.log.info("Starting message handler")

        while True:
            try:
                msg = await self.can_reader.get_message()

                if get_axis_id(msg) != self.axis_id:
                    # Ignore messages that aren't for this axis
                    continue

                db_msg = self.dbc.get_message_by_frame_id(msg.arbitration_id)

                if msg.is_remote_frame:
                    # RTR messages are requests for data, they don't have a data payload
                    self.log.debug(f"Get: {db_msg.name}")
                    # echo RTR messages back with data
                    self.send_message(db_msg.name)
                    continue

                # decode message data
                data = db_msg.decode(msg.data)
                cmd = db_msg.name.split("_", 1)[1]  # remove "AxisX_" prefix

                await self.execute_cmd(cmd, data)

            except KeyError:
                # If the message ID is not in the DBC file, print the raw message
                self.log.warning(f"Could not decode: {msg}")
            except Exception as e:
                self.log.error(f"Error: {e}")

    async def execute_cmd(self, cmd: str, data: dict):
        """execute command"""
        self.log.debug(f"Set: {cmd}: {data}")

        if cmd == "Set_Axis_State":
            await self.odrive.set_axis_state(data)
        elif cmd == "Set_Controller_Mode":
            self.odrive.set_controller_mode(data)
        elif cmd == "Set_Input_Pos":
            self.odrive.set_input_pos(data["Input_Pos"])
        elif cmd == "Set_Input_Vel":
            self.odrive.set_input_vel(data["Input_Vel"])

    def send_message(
        self, msg_name: str, msg_dict: Optional[dict] = None, rtr: bool = False
    ):
        """send message by name. If no msg_dict is provided, use zeros"""
        msg = self.dbc.get_message_by_name(msg_name)
        if rtr:
            # For RTR messages, don't specify the data field
            msg = can.Message(
                arbitration_id=msg.frame_id,
                is_extended_id=False,
                is_remote_frame=True,
            )
        else:
            full_msg_dict = {signal.name: 0 for signal in msg.signals}
            if msg_dict is not None:
                full_msg_dict.update(msg_dict)

            data = msg.encode(full_msg_dict)
            msg = can.Message(
                arbitration_id=msg.frame_id,
                data=data,
                is_extended_id=False,
            )

        self.bus.send(msg)  # type: ignore

    async def heartbeat_loop(self, delay: float = 0.2):
        """send heartbeat message"""
        self.log.info("Starting heartbeat loop")

        # Fetch the "Axis0_Heartbeat" message from the DBC database
        heartbeat_msg = self.dbc.get_message_by_name(f"Axis{self.axis_id}_Heartbeat")

        while True:
            # Construct the data payload using the DBC message definition
            data = heartbeat_msg.encode(
                {
                    "Axis_Error": 0,
                    "Axis_State": self.odrive.axis_state,
                    "Motor_Error_Flag": 0,
                    "Encoder_Error_Flag": 0,
                    "Controller_Error_Flag": 0,
                    "Trajectory_Done_Flag": 0,
                }
            )

            # Send the message
            message = can.Message(
                arbitration_id=heartbeat_msg.frame_id, data=data, is_extended_id=False
            )
            self.bus.send(message)

            await asyncio.sleep(delay)

    async def encoder_loop(self, delay: float = 0.1):
        """send encoder message"""
        self.log.info("Starting encoder loop")

        msg = self.dbc.get_message_by_name(f"Axis{self.axis_id}_Get_Encoder_Estimates")

        while True:
            self.odrive.update()
            data = msg.encode(
                {
                    "Pos_Estimate": self.odrive.position,
                    "Vel_Estimate": self.odrive.velocity,
                }
            )
            message = can.Message(
                arbitration_id=msg.frame_id, data=data, is_extended_id=False
            )
            self.bus.send(message)

            await asyncio.sleep(delay)

    async def main(self):
        """main loop"""

        async with asyncio.TaskGroup() as tg:
            tg.create_task(self.heartbeat_loop())
            tg.create_task(self.encoder_loop())
            tg.create_task(self.message_handler())

    def start(self):
        """start the main loop"""
        asyncio.run(self.main())

    def __del__(self):
        """destructor"""
        self.notifier.stop()
        self.bus.shutdown()


def main(axis_id: int = 0):
    try:
        mock = ODriveCANMock(axis_id)
        mock.start()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt")


if __name__ == "__main__":
    coloredlogs.install(level="DEBUG", fmt=LOG_FORMAT, datefmt=TIME_FORMAT)
    main()



# Example on how to use the DBC file to encode and decode messages
import can
from odrive_can import get_dbc
from odrive_can.odrive import CanMsg

db = get_dbc()  # load default DBC

# get message
db_msg = db.get_message_by_name("Axis0_Heartbeat")
print("Message:", db_msg)

# Both numeric and string values are accepted, choices are converted to integers
data = db_msg.encode(
    {
        "Axis_Error": "NONE",
        "Axis_State": 11,
        "Motor_Error_Flag": 0,
        "Encoder_Error_Flag": 0,
        "Controller_Error_Flag": 0,
        "Trajectory_Done_Flag": 0,
    }
)
print("Data:", data)

# decode message to dict
data_dict = db_msg.decode(data)
print("Decoded:", data_dict)
print("Frame ID:", db_msg.frame_id)

# get numeric value for "Axis_Error"
axis_error_value = data_dict["Axis_Error"]


# create CanMsg
msg = can.Message(arbitration_id=db_msg.frame_id, data=data)
print(f"{msg=}")

can_msg = CanMsg(msg)
print(f"{can_msg=}")



import asyncio
from odrive_can import ODriveCAN, CanMsg
from odrive_can.tools import UDP_Client

AXIS_ID = 0

SETPOINT = 50

udp = UDP_Client()  # send data to UDP server for plotting


def feedback_callback_fcn(msg: CanMsg, caller: ODriveCAN):
    """called on position estimate"""
    print(msg)
    udp.send(msg.data)


async def main():
    """connect to odrive"""
    drv = ODriveCAN(axis_id=AXIS_ID)

    # set up callback (optional)
    drv.feedback_callback = feedback_callback_fcn

    # start
    await drv.start()

    # check errors (raises exception if any)
    drv.check_errors()

    # set controller mode
    drv.set_controller_mode("POSITION_CONTROL", "POS_FILTER")

    # reset encoder
    drv.set_linear_count(0)

    # set axis state
    await drv.set_axis_state("CLOSED_LOOP_CONTROL")

    # set position gain
    drv.set_pos_gain(3.0)

    for _ in range(2):
        # setpoint
        drv.set_input_pos(SETPOINT)
        await asyncio.sleep(5.0)
        drv.set_input_pos(-SETPOINT)
        await asyncio.sleep(5.0)

    drv.set_input_pos(0.0)


asyncio.run(main())



#!/usr/bin/env python3
"""
 Inspect and decode ODrive CAN messages

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import odrive_can
from odrive_can.can_utils import get_can_bus

TIMEOUT = 1.0


def receive_and_decode(bus):
    dbc = odrive_can.get_dbc()
    nr_timeouts = 0
    idx = 0  # message index
    while True:
        msg = bus.recv(TIMEOUT)  # type: ignore

        if msg is None:
            print(".", end="", flush=True)
            nr_timeouts += 1
            continue
            # raise TimeoutError("Timeout occurred, no message.")

        # start new line if there were timeouts
        if nr_timeouts > 0:
            print("\n")
            nr_timeouts = 0

        axis_id = odrive_can.get_axis_id(msg)

        print(f"[{idx}] ", end="")

        if msg.is_remote_frame:
            # RTR messages are requests for data, they don't have a data payload
            db_msg = dbc.get_message_by_frame_id(msg.arbitration_id)
            print(f"Axis{axis_id} RTR: {db_msg.name}")
            continue

        try:
            # Attempt to decode the message using the DBC file
            db_msg = dbc.get_message_by_frame_id(msg.arbitration_id)
            decoded_message = db_msg.decode(
                msg.data
            )  # Remove msg.arbitration_id as it's not needed for decoding
            print(f"{db_msg.name}:{decoded_message}")
        except KeyError:
            # If the message ID is not in the DBC file, print the raw message
            print(f"Axis{axis_id} Raw Message: {msg}")

        idx += 1


def main():
    bus = get_can_bus()
    try:
        receive_and_decode(bus)
    except KeyboardInterrupt:
        print("Stopped")
    except TimeoutError as error:
        print(error)
    finally:
        bus.shutdown()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
 Timing utilities

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import asyncio
import time
import logging
from functools import wraps


def timeit(func):
    @wraps(func)
    async def async_wrapper(*args, **kwargs):
        logger = logging.getLogger(func.__module__)
        start_time = time.perf_counter()
        result = await func(*args, **kwargs)
        end_time = time.perf_counter()
        logger.info(
            f"Async function {func.__name__!r} executed in {end_time - start_time:.4f} seconds"
        )
        return result

    @wraps(func)
    def sync_wrapper(*args, **kwargs):
        logger = logging.getLogger(func.__module__)
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        logger.info(
            f"Function {func.__name__!r} executed in {end_time - start_time:.4f} seconds"
        )
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper


class Timer:
    """Timer class, including timeout"""

    def __init__(self, timeout):
        self.timeout = timeout
        self.start_time = time.time()

    def is_timeout(self):
        """Check if timeout has expired"""
        return time.time() - self.start_time > self.timeout

    def reset(self):
        """Reset timer"""
        self.start_time = time.time()

    def elapsed(self):
        """Return elapsed time since timer was started"""
        return time.time() - self.start_time



#!/usr/bin/env python3
"""
 Send data to UDP server, used for plotting with plotjuggler

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov


"""
import os
import time
import socket
import json
import logging
import math

# set this to the IP address of the machine running plotjuggler =


class UDP_Client:
    """send data to UDP server, used for plotting with plotjuggler

    Set the UDP_DATA_DEST environment variable to the IP address of the machine running plotjuggler

    """

    def __init__(self, host: str | None = None, port: int = 5005):

        if host is not None:
            self._host = host
        else:
            host = os.getenv("UDP_DATA_DEST", "localhost")
            self._host = host

        self._port = port

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self._log = logging.getLogger(self.__class__.__name__)
        self._log.info("UDP client created, sending to %s:%d", host, port)

    def send(self, data: dict, add_timestamp: bool = True):
        """send data to UDP server"""
        if add_timestamp:
            data["ts"] = time.time()

        try:
            self._sock.sendto(json.dumps(data).encode(), (self._host, self._port))
        except Exception as e:  # pylint: disable=broad-except
            self._log.error("Failed to send data: %s", e)


def demo(period: float = 2.0) -> None:
    """demo, send x and y values to plotjuggler
    x is sawtooth, y is sine wave"""
    client = UDP_Client()

    start_time = time.time()  # Record the start time
    while True:
        elapsed_time = time.time() - start_time  # Calculate elapsed time
        sine_value = math.sin(
            2 * math.pi * elapsed_time / period
        )  # Calculate sine value
        time.sleep(0.01)  # Small delay to control output frequency (adjust as needed)

        client.send({"x": elapsed_time % period, "y": sine_value})


if __name__ == "__main__":
    demo()



#!/usr/bin/env python3
"""
odrive_can CLI
"""

import json
import logging
import os
import subprocess
from pathlib import Path

import click
import coloredlogs  # type: ignore

from odrive_can import LOG_FORMAT, TIME_FORMAT
from odrive_can.utils import run_main_async

log = logging.getLogger()
coloredlogs.install(level="INFO", fmt=LOG_FORMAT, datefmt=TIME_FORMAT)

# pylint: disable=import-outside-toplevel, unused-argument, broad-except

# ------------------ helpers


@click.group()
def cli():
    pass  # pragma: no cover


@cli.group()
def demo():
    """demonstration of control modes"""


@cli.command()
def info():
    """Print package info"""
    from odrive_can import __version__

    print(__version__)


@cli.command()
@click.option("--axis-id", default=0, help="ODrive axis ID")
def mock(axis_id):
    """Mock ODrive CAN interface"""
    from .mock import main

    main(axis_id=axis_id)


@cli.command()
def inspect():
    """Inspect and decode ODrive CAN messages"""
    from .inspector import main

    main()


@cli.command()
@click.option(
    "--output-file",
    default="odrive_config.json",
    help='The name of the output file (default: "odrive_config.json")',
)
def backup(output_file):
    """Backup config to config folder"""
    tmp_path = Path("/tmp")
    odrive_files = tmp_path.glob("odrive*")

    # 1. Remove all /tmp/odrive* files
    for file in odrive_files:
        file.unlink()
        print(f"Removed: {file}")

    # 2. Run `odrivetool backup-config`
    try:
        result = subprocess.run(
            "odrivetool backup-config",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
            timeout=5,
        )
        if result.returncode != 0:
            print("Failed to run 'odrivetool backup-config'")
            print(result.stderr.decode())
            return
    except subprocess.TimeoutExpired:
        print("Failed to run 'odrivetool backup-config'")
        print("Timeout expired")
        return

    # 3. Find the generated `odrive-config-*.json` file and read it
    config_files = list(tmp_path.glob("odrive-config-*.json"))
    if not config_files:
        print("No config file found")
        return

    config_file = max(config_files, key=os.path.getctime)  # Get the latest file
    data = config_file.read_text()

    # 4. Save parsed data using json, with human readable formatting
    parsed_data = json.loads(data)

    # Save the formatted data to the 'config' directory
    dest = Path(output_file)
    with dest.open("w", encoding="utf8") as f:
        json.dump(parsed_data, f, indent=2)
    print(f"Formatted config saved to {dest}")


@demo.command()
@click.option("--axis-id", default=1, help="ODrive axis ID")
@click.option("--amplitude", default=40, help="Amplitude")
@click.option(
    "--input-mode",
    type=click.Choice(["POS_FILTER", "TRAP_TRAJ"], case_sensitive=False),
    default="POS_FILTER",
    help="Input mode (POS_FILTER or TRAP_TRAJ)",
)
def position(axis_id, input_mode, amplitude):
    """Position control demo"""
    from .examples.position_control import main

    try:
        run_main_async(main(axis_id, input_mode, amplitude))
    except Exception as e:
        log.error(e)


@demo.command()
@click.option("--axis-id", default=1, help="ODrive axis ID")
@click.option("--amplitude", default=40, help="Amplitude")
def velocity(axis_id, amplitude):
    """Velocity control demo"""
    from .examples.velocity_control import main

    try:
        run_main_async(main(axis_id, amplitude))
    except Exception as e:
        log.error(e)


@demo.command()
@click.option("--axis-id", default=1, help="ODrive axis ID")
@click.option("--amplitude", default=40, help="Amplitude")
def watchdog(axis_id, amplitude):
    """Demonstrate the watchdog feature"""
    from .examples.watchdog import main

    try:
        run_main_async(main(axis_id, amplitude))
    except Exception as e:
        log.error(e)


if __name__ == "__main__":
    cli()  # pragma: no cover



#!/usr/bin/env python3
"""
 utils for odrive testing

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""


class OdriveError(Exception):
    """custom exception"""


def check_error(drv):
    """check if there are any errors"""

    ax = drv.axis0

    errors = [
        ("ax.error", ax.error),
        ("ax.motor.error", ax.motor.error),
        ("ax.encoder.error", ax.encoder.error),
        ("ax.controller.error", ax.controller.error),
        ("ax.sensorless_estimator.error", ax.sensorless_estimator.error),
    ]

    for var_name, var_value in errors:
        if var_value > 0:
            raise OdriveError(f"Error in {var_name}: {var_value}")



#!/usr/bin/env python3
"""
 simple acceleration test

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import time
import asyncio
import odrive  # type: ignore
import odrive.enums as enums  # type: ignore
from utils import check_error

from odrive_can.tools import UDP_Client


RAMP_RATE = 200
SETPOINT = 40

drv = odrive.find_any()

print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()
drv.axis0.requested_state = enums.AxisState.CLOSED_LOOP_CONTROL

check_error(drv)

# configure velocity control
drv.axis0.controller.config.input_mode = enums.InputMode.VEL_RAMP
drv.axis0.controller.config.control_mode = enums.ControlMode.VELOCITY_CONTROL
drv.axis0.controller.config.vel_ramp_rate = RAMP_RATE  # this only works with VEL_RAMP
drv.axis0.controller.config.vel_limit = 150.0
drv.axis0.motor.config.current_lim = 10.0

# check correct setti


async def feedback_loop():
    ax = drv.axis0
    udp = UDP_Client()

    while True:
        check_error(drv)

        data = {
            "ts": time.time(),
            "sp": ax.controller.input_vel,
            "vel": ax.encoder.vel_estimate,
            "pos": ax.encoder.pos_estimate,
        }

        udp.send(data)
        await asyncio.sleep(0.05)


async def setpoint_loop():
    while True:
        drv.axis0.controller.input_vel = SETPOINT
        await asyncio.sleep(2)
        drv.axis0.controller.input_vel = -SETPOINT
        await asyncio.sleep(2)


async def main():
    await asyncio.gather(feedback_loop(), setpoint_loop())


try:
    asyncio.run(main())
except KeyboardInterrupt:
    pass
finally:
    drv.axis0.controller.input_vel = 0
    drv.axis0.requested_state = enums.AxisState.IDLE



#!/usr/bin/env python3
"""
 Generate documentation from dbc file

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

from odrive_can import get_dbc


def dbc_to_markdown(db, markdown_file_path, name):
    """create documentation from dbc database"""

    print(f"Generating {markdown_file_path}")
    # Open the Markdown file for writing
    with open(markdown_file_path, "w", encoding="utf8") as md_file:
        # Write a header
        md_file.write(f"## {name}  interface\n\n")

        # Iterate over messages
        for message in db.messages:
            md_file.write(f"### ID: {message.frame_id} - {message.name}\n")
            md_file.write(f"- Name: {message.name}\n")
            md_file.write(f"- Length: {message.length} bytes\n")
            md_file.write(f"- Sender: {message.senders}\n\n")

            if message.frame_id <= 29:
                # Write a table header for signals
                md_file.write(
                    "| Signal Name | Start Bit | Length | Factor | Offset | Min Value | Max Value | Unit | Receiver | Choices |\n"
                )
                md_file.write(
                    "|-------------|-----------|--------|--------|--------|-----------|-----------|------|----------|---------|\n"
                )

                # Iterate over signals
                for signal in message.signals:
                    choices_str = (
                        ", ".join(f"{k}: {v}" for k, v in signal.choices.items())
                        if signal.choices
                        else ""
                    )
                    md_file.write(
                        f"| {signal.name} | {signal.start} | {signal.length} | {signal.scale} | {signal.offset} | {signal.minimum} | {signal.maximum} | {signal.unit or ''} | {', '.join(signal.receivers)} | {choices_str} |\n"
                    )

                # Add a newline after each message
                md_file.write("\n")


name = "odrive-cansimple-0.5.6"
db = get_dbc(name)
dbc_to_markdown(db, f"{name}.md", name)



#!/usr/bin/env python3
"""
 simple acceleration test

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import asyncio
import odrive  # type: ignore
import odrive.enums as enums  # type: ignore
from utils import check_error

from odrive_can.tools import UDP_Client

DURATION = 8
SETPOINT = 120

drv = odrive.find_any()
ax = drv.axis0


print(f"Found ODrive {drv.serial_number}")

drv.clear_errors()

check_error(drv)
ax.controller.input_pos = 0
ax.encoder.set_linear_count(0)

ax.controller.config.control_mode = enums.ControlMode.POSITION_CONTROL
ax.controller.config.input_mode = enums.InputMode.TRAP_TRAJ
ax.requested_state = enums.AxisState.CLOSED_LOOP_CONTROL
ax.controller.config.input_filter_bandwidth = 4.0
ax.motor.config.current_lim = 5.0
ax.controller.config.vel_limit = 120.0


# trajectory control
ax.trap_traj.config.vel_limit = 50.0
ax.trap_traj.config.accel_limit = 40.0
ax.trap_traj.config.decel_limit = 40.0


# position control
ax.controller.config.pos_gain = 0.5


async def feedback_loop():
    udp = UDP_Client()

    while True:
        check_error(drv)

        data = {
            "sp": ax.controller.input_pos,
            "vel": ax.encoder.vel_estimate,
            "pos": ax.encoder.pos_estimate,
        }

        udp.send(data)
        await asyncio.sleep(0.05)


async def setpoint_loop():
    # set positiion to zero

    while True:
        ax.controller.input_pos = SETPOINT
        await asyncio.sleep(DURATION)
        ax.controller.input_pos = -SETPOINT
        await asyncio.sleep(DURATION)


async def main():
    await asyncio.gather(feedback_loop(), setpoint_loop())


try:
    asyncio.run(main())
except KeyboardInterrupt:
    pass
finally:
    ax.requested_state = enums.AxisState.IDLE
